package debug;

import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.pdmodel.PDPageContentStream;
import org.apache.pdfbox.pdmodel.common.PDRectangle;
import org.apache.pdfbox.pdmodel.font.PDFont;
import org.apache.pdfbox.pdmodel.font.PDType1Font;

import java.awt.*;
import java.io.IOException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * <p>Specification is on first page of attachment "Protokoll.pdf" at https://jira.manz.at/browse/ERVSW-442.</p>
 * <br>
 * Page header: <i>"Protokoll vom 02.04.2019"</i>, horizontal line below.<br>
 * Table title: <i>"Ausgang Hinverkehr (3)"</i>,  3 = number of table rows<br>
 * Table column headers:
 * <blockquote>
 *      <i>"Erstellt Eingebracht Schriftsatzart Aktenzeichen Einbringer Status"</i>,<br>
 * </blockquote>
 * Table title: <i>"Eingang Rückverkehr (3)"</i>,  3 = Anzahl der Tabellenzeilen<br>
 * Table column headers:
 * <blockquote>
 *      <i>"Hinterlegt Abgerufen Schriftsatzart Aktenzeichen Einbringer Status"</i>
 * </blockquote>
 * Page footer: <i>"MANZ webERV: Vorname Nachname       Seite 1/1"</i>, left is name of webERV user, right aligned page number 1-n / total sum of pages<br>
 */
public class JournalPdf {

    private static final int COLUMN_COUNT = 6;
    private static final float[] COLUMN_PROPORTIONS = new float[] { 75f, 75f, 140f, 150f, 80f, 80f };   // taken from old webERV application

    private static String[][] schriftsaetzeHeaderTableData = new String[][] {
        new String [] {
                "Erstellt",
                "Eingebracht",
                "Schriftsatzart",
                "Aktenzeichen",
                "Einbringer",
                "Status"
        }
    };

    private static String[][] erledigungenHeaderTableData = new String[][] {
            new String [] {
                "Hinterlegt",
                "Abgerufen",
                "Schriftsatzart",
                "Aktenzeichen",
                "Einbringer",
                "Status"
            }
    };

    private static final DateTimeFormatter DATEFORMATTER = DateTimeFormatter.ofPattern("dd.MM.yyyy");


    private final PDDocument document;

    private final float pageMarginLeft = 20f;
    private final float horizontalLinePadding = 2f;
    private final float horizontalLineThickness = 1f;
    private final float verticalSpaceAfterHeadline = 8f;
    private final float verticalSpaceAfterTable = verticalSpaceAfterHeadline;

    private float pageWidth;
    private float pageHeight;
    private float headerHeight;
    private float footerHeight;
    private float tableHeaderHeight;
    private float[] tableColumnWidths;

    private PdfTextWriter textWriter;

    private boolean geometryCalculation;
    private Set<SchriftverkehrResultItem> allItemsWhereToBreakPageBefore;


    public JournalPdf(PDDocument document)  throws IOException   {
        assert COLUMN_COUNT == schriftsaetzeHeaderTableData[0].length && COLUMN_COUNT == erledigungenHeaderTableData[0].length;
        assert document != null;

        this.document = document;
    }

    /**
     * Prints (1) given schriftsaetze, (2) given erledigungen, each of them or both
     * together could span several pages, or they could fit together on just one page.
     * @param schriftsaetze items to print first.
     * @param erledigungen items to print second.
     * @throws IOException coming from PdfBox.
     */
    public void print(List<SchriftverkehrResultItem> schriftsaetze, List<SchriftverkehrResultItem> erledigungen) throws IOException  {
        assert schriftsaetze != null && erledigungen != null;

        newPage();
        assert pageHeight > 0 && pageWidth > 0;

        tableColumnWidths = PdfBoxUtil.adaptColumnWidths(COLUMN_PROPORTIONS, pageWidth - (2 * pageMarginLeft));
        assert tableColumnWidths.length == COLUMN_COUNT;

        final List<List<SchriftverkehrResultItem>> schriftsaetzePageParts = new ArrayList<>();  // a "page part" can be a page or just a part of a page,
        final List<List<SchriftverkehrResultItem>> erledigungenPageParts = new ArrayList<>();   // two page parts may be on same page
        calculatePageParts(schriftsaetze, schriftsaetzePageParts, erledigungen, erledigungenPageParts);

        // output
        float height = pageHeight - headerHeight;  // start at top of first page
        height = printTablePagePartsWithTitle(height, true, schriftsaetze.size(), schriftsaetzePageParts);
        height = printTablePagePartsWithTitle(height, false, erledigungen.size(), erledigungenPageParts);

        textWriter.close();
        textWriter = null;

        addHeaderAndFooterOnAllPages();
    }


    private void newPage() throws IOException   {
        assert geometryCalculation == false;

        if (textWriter != null)
            textWriter.close();

        final PDPage page = new PDPage(PDRectangle.A4);
        document.addPage(page);
        pageWidth = page.getMediaBox().getWidth();
        pageHeight = page.getMediaBox().getHeight();

        textWriter = newTextWriter(page);
    }

    private PdfTextWriter newTextWriter(PDPage page) throws IOException   {
        final PDFont font = PDType1Font.TIMES_ROMAN;
        final float fontSize = 10f;
        return new PdfTextWriter(new PDPageContentStream(document, page, PDPageContentStream.AppendMode.APPEND, false), font, fontSize);
    }


    // START layout calculation

    private void calculatePageParts(
            List<SchriftverkehrResultItem> schriftsaetze,
            List<List<SchriftverkehrResultItem>> schriftsaetzePageParts,
            List<SchriftverkehrResultItem> erledigungen,
            List<List<SchriftverkehrResultItem>> erledigungenPageParts) throws IOException
    {
        // for height calculation, first set a writer that doesn't really write
        final PdfTextWriter originalTextWriter = this.textWriter;
        geometryCalculation = true;
        try {
            this.textWriter = new PdfTextWriter(null, originalTextWriter.getFont(), originalTextWriter.getFontSize());  // writer for header- and footer height
            calculateHeaderAndFooterHeights();
            final float calculationPageHeight = pageHeight - headerHeight - footerHeight - 5f;

            final float START_Y = 0f;
            float calculationY = START_Y;
            this.textWriter = new PageCountCalculationTextWriter(calculationY, calculationPageHeight);
            calculationY = splitIntoPageParts(calculationY, calculationPageHeight, true, schriftsaetze, schriftsaetzePageParts);
            splitIntoPageParts(calculationY, calculationPageHeight, false, erledigungen, erledigungenPageParts);
        }
        finally {   // reset calculation mode
            geometryCalculation = false;
            textWriter = originalTextWriter;
        }
    }

    private void calculateHeaderAndFooterHeights() throws IOException {
        float height = drawHeader(0f);    // goes into minus
        headerHeight = (-height);

        height = drawFooter(0f, 0, 0);    // goes into minus
        footerHeight = (-height);

        height = drawTableHeader(0f, new String[][] { new String[] { "Dummy" } }, new float[] { 20f }, null);    // goes into minus
        tableHeaderHeight = (-height);

        assert headerHeight > 0f && footerHeight > 0f && tableHeaderHeight > 0f;
    }


    private class PageCountCalculationTextWriter extends PdfTextWriter
    {
        final Set<SchriftverkehrResultItem> itemsWhereToBreakPageBefore = new HashSet<>();
        private final float pageHeightWithoutHeaderAndFooter;
        private float bottomPageBoundary;

        PageCountCalculationTextWriter(final float startY, final float pageHeightWithoutHeaderAndFooter) throws IOException {
            super(null, textWriter.getFont(), textWriter.getFontSize());    // null: run without writing
            
            this.pageHeightWithoutHeaderAndFooter = pageHeightWithoutHeaderAndFooter;
            assert this.pageHeightWithoutHeaderAndFooter > 0f;
            bottomPageBoundary = startY - this.pageHeightWithoutHeaderAndFooter;
        }
        
        @Override
        public void advancedVertically(float newHeight, Object item) {
            final int newHeightInt = (int) Math.ceil(newHeight);

            if (exceedsCurrentPage(newHeightInt))  {   // exceeded bottom, put item into next page
                itemsWhereToBreakPageBefore.add((SchriftverkehrResultItem) item);
                System.err.println("Exceed to new page at item "+item+", currentHeight="+bottomPageBoundary+", newHeight="+newHeightInt+", at "+item);
                
                bottomPageBoundary -= pageHeightWithoutHeaderAndFooter;
                // because this item gets shifted to next page, there will be a table header too, 
                // but during pagination the header is not painted, so make page a little smaller
                bottomPageBoundary += tableHeaderHeight;
            }
        }

        private boolean exceedsCurrentPage(int newHeight) {
            return bottomPageBoundary >= newHeight;
        }
    }


    private float splitIntoPageParts(
            float height,
            float calculationPageHeight,
            boolean isSchriftsaetze,
            List<SchriftverkehrResultItem> items,
            final List<List<SchriftverkehrResultItem>> resultingPages)
        throws IOException
    {
        assert resultingPages != null && resultingPages.size() <= 0;

        final List<List<SchriftverkehrResultItem>> calculationPages = new ArrayList<>();
        calculationPages.add(items);    // for calculation: all in one page-part
        height = printTablePagePartsWithTitle(height, isSchriftsaetze, items.size(), calculationPages);

        final Set<SchriftverkehrResultItem> itemsWhereToBreakPageBefore = ((PageCountCalculationTextWriter) textWriter).itemsWhereToBreakPageBefore;
        assemblePageParts(items, itemsWhereToBreakPageBefore, resultingPages);

        return height;
    }

    private void assemblePageParts(List<SchriftverkehrResultItem> items, Set<SchriftverkehrResultItem> itemsWhereToBreakPageBefore, List<List<SchriftverkehrResultItem>> resultingPages) {
        if (allItemsWhereToBreakPageBefore == null)
            allItemsWhereToBreakPageBefore = new HashSet<SchriftverkehrResultItem>();

        allItemsWhereToBreakPageBefore.addAll(itemsWhereToBreakPageBefore);

        for (SchriftverkehrResultItem item : items) {
            if (itemsWhereToBreakPageBefore.contains(item)) {
                final List<SchriftverkehrResultItem> nextPagePartItems = new ArrayList<>();
                resultingPages.add(nextPagePartItems);
                nextPagePartItems.add(item);
            }
            else    {
                final List<SchriftverkehrResultItem> currentPagePartItems;
                if (resultingPages.size() <= 0)
                    resultingPages.add(currentPagePartItems = new ArrayList<>());
                else
                    currentPagePartItems = resultingPages.get(resultingPages.size() - 1);

                currentPagePartItems.add(item);
            }
        }
    }

    // END layout calculation


    private float printTablePagePartsWithTitle(float height, boolean isSchriftsaetze, int itemsCount, List<List<SchriftverkehrResultItem>> pageParts) throws IOException {
        if (pageParts.size() <= 0 || pageParts.get(0).size() <= 0)
            return height;
        
        // take title to next page when table would start on next page
        final SchriftverkehrResultItem firstItemOfFirstPagePart = pageParts.get(0).get(0);
        height = newPageWhenPartBreaksPage(height, firstItemOfFirstPagePart);
        // draw title
        height = drawTitle(height, isSchriftsaetze ? "Ausgang Hinverkehr" : "Eingang Rückverkehr", itemsCount);

        for (List<SchriftverkehrResultItem> pagePartItems : pageParts) {
            final SchriftverkehrResultItem firstItemOfPagePart = pagePartItems.get(0);
            if (firstItemOfPagePart != firstItemOfFirstPagePart)    // don't repeat if a page break already happened
                height = newPageWhenPartBreaksPage(height, firstItemOfPagePart);

            // draw table
            height = printTable(height, pagePartItems, isSchriftsaetze);
        }
        return height;
    }

    private float newPageWhenPartBreaksPage(float height, SchriftverkehrResultItem pagePartItem) throws IOException {
        if (geometryCalculation == false && allItemsWhereToBreakPageBefore.contains(pagePartItem))   {
            newPage();
            height = pageHeight - headerHeight;  // prepare next page
        }
        return height;
    }

    private float printTable(float height, List<SchriftverkehrResultItem> items, boolean isSchriftsaetze) throws IOException  {
        final String[][] headerTableData = isSchriftsaetze ? schriftsaetzeHeaderTableData : erledigungenHeaderTableData;
        final String[][] cellData = isSchriftsaetze ? toSchriftsaetzeTableData(items) : toErledigungenTableData(items);
        return drawTable(height, headerTableData, cellData, tableColumnWidths, items);
    }

    private float drawTableHeader(float height, String[][] headerTableData, float[] columnWidths, SchriftverkehrResultItem firstItem) throws IOException {
        final List<SchriftverkehrResultItem> headerItemList = new ArrayList<>();
        headerItemList.add(firstItem);
        height = PdfBoxUtil.drawTable(textWriter, pageMarginLeft, height, headerTableData, columnWidths, headerItemList);

        height = drawHorizontalLine(height, Color.LIGHT_GRAY);

        return height;
    }

    private float drawTable(float height, String[][] headerTableData, String[][] cellData, float[] columnWidths, List<SchriftverkehrResultItem> items) throws IOException {
        // draw header as separate table
        height = drawTableHeader(height, headerTableData, columnWidths, (items.size() > 0) ? items.get(0) : null);

        // draw table
        height = PdfBoxUtil.drawTable(textWriter, pageMarginLeft, height, cellData, columnWidths, items);
        
        if (geometryCalculation == false)
            assert height >= 0;
        
        height -= verticalSpaceAfterTable;

        return height;
    }

    private float drawHorizontalLine(float height, Color color) throws IOException {
        height -= horizontalLinePadding;
        PdfBoxUtil.drawLine(textWriter, pageMarginLeft, height, pageWidth - pageMarginLeft, height, color, horizontalLineThickness);
        height -= horizontalLineThickness;
        height -= horizontalLinePadding;

        return height;
    }

    private float drawText(float height, float marginLeft, String text) throws IOException {
        PdfBoxUtil.drawText(textWriter, marginLeft, height, text);
        height -= textWriter.getTextLineHeight();
        
        return height;
    }

    private float drawTitle(float height, String title, int numberOfItems) throws IOException {
        final String schriftsaetzeTitle = title+" ("+numberOfItems+")";
        textWriter.setFont(PDType1Font.TIMES_BOLD);
        height -= textWriter.getTextLineHeight();
        height = drawText(height, pageMarginLeft, schriftsaetzeTitle);
        textWriter.setFont(PDType1Font.TIMES_ROMAN);
        height -= verticalSpaceAfterHeadline;   // some vertical space after title

        return height;
    }

    private float drawHeader(float height) throws IOException {
        final String topTitle = "Protokoll vom "+format(LocalDateTime.now());

        height -= textWriter.getTextLineHeight();   // go to baseline of text
        drawText(height, pageMarginLeft, topTitle); // don't assign height, it was already calculated above
        height -= horizontalLinePadding;

        height = drawHorizontalLine(height, Color.GRAY);
        height -= verticalSpaceAfterHeadline;

        return height;
    }

    private float drawFooter(float height, int pageNumber, int totalPages) throws IOException {
        final String bottomTitleLeft = "MANZ webERV: Vorname Nachname"; // TODO: put login-user names there
        final String bottomTitleRight = "Seite "+pageNumber+"/"+totalPages;

        height = drawHorizontalLine(height, Color.GRAY);
        height -= textWriter.getFontHeight();   // just go to bottom-baseline of text, padding to line has already been subtracted

        drawText(height, pageMarginLeft, bottomTitleLeft);  // don't assign height because next string goes into same line
        return drawText(height, pageWidth - pageMarginLeft - textWriter.getStringWidth(bottomTitleRight), bottomTitleRight);  // right aligned
    }

    private void addHeaderAndFooterOnAllPages() throws IOException {
        final int totalPages = document.getNumberOfPages();
        for (int i = 0; i < document.getNumberOfPages(); i++)   {
            final PDPage page = document.getPage(i);
            this.textWriter = newTextWriter(page);
            drawHeader(pageHeight);
            drawFooter(footerHeight, i + 1, totalPages);
            textWriter.close();
        }
    }


    private String[][] toSchriftsaetzeTableData(List<SchriftverkehrResultItem> schriftsaetze) throws IOException {
        final int schriftsaetzeRowCount = (int) schriftsaetze.size();
        final String[][] schriftsaetzeCellData = new String[schriftsaetzeRowCount][COLUMN_COUNT];

        for (int i = 0; i < schriftsaetzeRowCount; i++) {
            final SchriftverkehrResultItem item = schriftsaetze.get(i);
            String[] row = new String[COLUMN_COUNT];
            schriftsaetzeCellData[i] = row;
            row[0] = format(item.getErstellungsZeitpunkt());
            row[1] = format(item.getEinbringungsZeitpunkt());
            row[2] = item.getInhaltArtLabel();
            row[3] = item.getAktenzeichen();
            row[4] = item.getEinbringerLabel();
            row[5] = item.getStatus();
        }
        return schriftsaetzeCellData;
    }

    private String[][] toErledigungenTableData(List<SchriftverkehrResultItem> erledigungen) throws IOException {
        final int erledigungenRowCount = (int) erledigungen.size();
        final String[][] erledigungenCellData = new String[erledigungenRowCount][COLUMN_COUNT];

        for (int i = 0; i < erledigungenRowCount; i++)  {
            final SchriftverkehrResultItem item = erledigungen.get(i);
            String[] row = new String[COLUMN_COUNT];
            erledigungenCellData[i] = row;
            row[0] = format(item.getVersandZeitpunkt());
            row[1] = format(item.getEinbringungsZeitpunkt());
            row[2] = item.getInhaltArtLabel();
            row[3] = item.getAktenzeichen();
            row[4] = item.getEinbringerLabel();
            row[5] = item.getStatus();
        }
        return erledigungenCellData;
    }

    private String format(LocalDateTime date)    {
        return (date == null) ? "" : DATEFORMATTER.format(date);
    }

}
