package debug;

import org.apache.pdfbox.pdmodel.font.PDFont;

import java.awt.*;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

public final class PdfBoxUtil {

    /**
     * PDF table adapted from https://stackoverflow.com/questions/3871879/apache-pdfbox-java-library-is-there-an-api-for-creating-tables.
     * Draws given text matrix using given column widths. Long texts inside cells get broken
     * down into several lines, which increases the table row height where that cell resides.
     * @param textWriter the writer to use for drawing into the current PDF page.
     * @param startX left-upper horizontal coordinate, or page margin left.
     * @param startY left-upper vertical coordinate, or page margin top. Mind that y=0 is on bottom of page!
     * @param content matrix of cell strings for the table.
     * @param columnWidths the widths of the columns of the table, left-to-right.
     * @param rowItems the list of original row-objects from which content was extracted.
     * @return the vertical coordinate where the table ends.
     */
    public static float drawTable(
            PdfTextWriter textWriter,
            final float startX,
            final float startY,
            final String[][] content,
            final float [] columnWidths,
            List<? extends Object> rowItems) throws IOException
    {
        assert content.length == rowItems.size();
        assert columnWidths != null && content[0].length == columnWidths.length;

        // draw cell texts, without grid
        float x = startX;
        float y = startY;
        final int rows = content.length;

        for (int row = 0; row < rows; row++)    {
            final float columns = content[row].length;
            float minY = Float.MAX_VALUE;

            for (int column = 0 ; column < columns; column++) {
                float newY = PdfBoxUtil.drawTextIntoRectangle(textWriter, content[row][column], columnWidths[column], x, y, rowItems.size() > row ? rowItems.get(row) : null);

                if (newY < minY)
                    minY = newY;

                x += columnWidths[column];
            }
            y = minY;
            x = startX;
        }

        return y;
    }

    public static void drawLine(PdfTextWriter textWriter, float xStart, float yStart, float xEnd, float yEnd, Color color, float lineThickness) throws IOException {
        textWriter.setStrokingColor(color);
        textWriter.setLineWidth(lineThickness);
        textWriter.moveTo(xStart, yStart);
        textWriter.lineTo(xEnd, yEnd);
        textWriter.stroke();
        textWriter.setStrokingColor(Color.BLACK);
    }

    public static void drawText(PdfTextWriter textWriter, float x, float y, String text) throws IOException {
        textWriter.beginText();
        textWriter.newLineAtOffset(x, y);
        textWriter.showText(text);
        textWriter.endText();
    }

    /**
     * Calculates new column widths to match given table width proportionally.
     * Does NOT modify given column widths!
     * @param columnWidths the intended widths of columns of the target table.
     * @param tableWidth the entire width of the target table.
     * @return new array containing corrected column widths.
     */
    public static float[] adaptColumnWidths(float[] columnWidths, float tableWidth) {
        float sum = 0.0f;
        for (int i = 0; i < columnWidths.length; i++)
            sum += columnWidths[i];

        final float proportion = tableWidth / sum;

        float[] newColumnWidths = new float[columnWidths.length];
        for (int i = 0; i < columnWidths.length; i++)
            newColumnWidths[i] = proportion * columnWidths[i];

        return newColumnWidths;
    }


    /** @return the width of given single-line text (newlines are not considered). */
    public static float stringWidth(PDFont font, float fontSize, String text) throws IOException  {
        return font.getStringWidth(text) * fontSize / 1000f;
    }

    /** @return the height of given font, which is different from PdfBox fontSize. */
    public static float fontHeight(PDFont font, float fontSize) throws IOException  {
        return fontSize * font.getFontDescriptor().getCapHeight() / 1000f;
    }

    /** @return a recommended line-height, which is of 150% of the font-height (like web-browsers do). */
    public static float lineHeight(PDFont font, float fontSize) throws IOException  {
        return fontHeight(font, fontSize) * 1.5f;
    }

    /** @return the height of the font above lower-case letter upper boundary, i.e. the additional height that a 'T' would consume. */
    public static float ascent(PDFont font, float fontSize) throws IOException  {
        return fontSize * font.getFontDescriptor().getAscent() / 1000f;
    }

    /** @return the (positive) depth of the font below baseline, i.e. the height that a 'g' would consume below its writing line. */
    public static float descent(PDFont font, float fontSize) throws IOException  {
        return fontSize * -font.getFontDescriptor().getDescent() / 1000f;
    }


    /**
     * Draws given single- or multi-line text into an open-bottom rectangle given by startX, startY and width.
     * @param textWriter holding the content-stream (and font-settings) where to draw into.
     * @param text the single- or multi-line text to0 draw.
     * @param width the width of the open-bottom rectangle where to draw into.
     * @param startX the left coordinate of the rectangle.
     * @param startY the top coordinate of the rectangle.
     * @param rowItem the row-object from which text was extracted, will be passed to advancedVertically().
     * @return the height where writing stopped, i.e. the bottom line of the text, including font descent.
     * @throws IOException when PdfBox throws one.
     */
    public static float drawTextIntoRectangle(
            PdfTextWriter textWriter,
            String text,
            float width,
            final float startX,
            final float startY,
            Object rowItem) throws IOException
    {
        final float lineHeight = textWriter.getTextLineHeight();
        
        float y = startY;
        y -= lineHeight;   // go down to text baseline

        textWriter.beginText();
        textWriter.newLineAtOffset(startX, y);
        final List<String> lines = splitToFitWidth(text, width, textWriter);
        for (String line : lines)    {
            
            textWriter.advancedVertically(y, rowItem);
            
            textWriter.showText(line);
            if (line != lines.get(lines.size() - 1))    {
                textWriter.newLineAtOffset(0, -lineHeight);
                y -= lineHeight;
            }
            else    {
                y -= descent(textWriter.getFont(), textWriter.getFontSize());
            }
        }
        textWriter.endText();
        System.err.println("Consumed "+(startY - y)+" at "+y+" for "+rowItem);
        
        return y;
    }

    private static List<String> splitToFitWidth(String text, final float width, PdfTextWriter textWriter) throws IOException {
        text = removeUnreadableCharacters(text); // PdfBox doesn't accept tabs
        text = text.replaceAll("[ ]{2,}", " "); // normalize spaces

        String[] paragraphs = text.split("\r\n");   // WINDOWS
        if (paragraphs.length == 1)
            paragraphs = text.split("\n");  // UNIX, OSX

        final List<String> lines = new ArrayList<String>();
        for (String paragraph : paragraphs)
            for (String line : splitParagraphToFitWidth(paragraph, width, textWriter))
                lines.add(line);

        return lines;
    }

    private static List<String> splitParagraphToFitWidth(String text, final float width, PdfTextWriter textWriter) throws IOException {
        final List<String> lines = new ArrayList<String>();
        if (text.length() <= 0) {   // preserve newlines
            lines.add("");
            return lines;
        }

        int previousSpaceIndex = -1;

        while (text.length() > 0) {
            int nextSpaceIndex = nextSpaceIndex(text, previousSpaceIndex + 1);  // search first space
            if (nextSpaceIndex < 0)
                nextSpaceIndex = text.length(); // no space found

            final String line = text.substring(0, nextSpaceIndex).trim();
            final float lineWidth = textWriter.getStringWidth(line);

            if (lineWidth > width) {   // must split at preceding space if any
                if (previousSpaceIndex < 0) // no preceding space, split by width
                    previousSpaceIndex = nextSplitIndex(text, width, textWriter);

                lines.add(text.substring(0, previousSpaceIndex));
                text = text.substring(previousSpaceIndex).trim();
                previousSpaceIndex = -1;
            }
            else if (nextSpaceIndex == text.length()) { // reached end of text
                lines.add(line);
                text = "";
            }
            else    {
                previousSpaceIndex = nextSpaceIndex;
            }
        }

        return lines;
    }

    private static int nextSpaceIndex(String text, int searchStartIndex) {
        for (int i = searchStartIndex; i < text.length(); i++)  {
            char c = text.charAt(i);

            if (Character.isWhitespace(c))
                return i;   // this trailing space will be removed

            if (i + 1 < text.length() && (c == '-' || c == '_'))
                return i + 1;   // break AFTER character
        }
        return -1;  // not found
    }

    private static int nextSplitIndex(String text, float width, PdfTextWriter textWriter) throws IOException {
        width -= textWriter.getFontSize() * 0.39f;   // empiric correction
        for (int i = 1; i < text.length(); i++) {
            final String line = text.substring(0, i).trim();
            final float lineWidth = textWriter.getStringWidth(line);
            if (lineWidth >= width)
                return i;
        }
        return text.length();   // not found
    }

    private static String removeUnreadableCharacters(String text)  {
        return text.replaceAll("\t", " "); // PdfBox doesn't accept tabs
    }
}

// Following code can be used to draw table/row/cell borders in drawTable()

//        final int columns = content[0].length;
//        final float tableHeight = rowHeight * rows;
//
//        // draw border lines
//        float nexty = y ;
//        for (int i = 0; i <= rows; i++) {
//            drawLine(contentStream, x, nexty, x + tableWidth, nexty);
//            nexty -= rowHeight;
//        }
//
//        // draw the column border lines
//        float nextx = x;
//        for (int i = 0; i <= columns; i++) {
//            drawLine(contentStream, nextx, y, nextx, y - tableHeight);
//            if (i < columnWidths.length)
//                nextx += columnWidths[i];
//        }



//        float textx = x + cellPaddingLeftBottom;
//        float texty = y - rowHeight + cellPaddingLeftBottom;
//        for (int row = 0; row < rows; row++){
//            final float columns = content[row].length;
//            for (int column = 0 ; column < columns; column++) {
//                drawText(contentStream, textx, texty, content[row][column]);
//                textx += columnWidths[column];
//            }
//            texty -= rowHeight;
//            textx = x + cellPaddingLeftBottom;
//        }
//        return y - (rowHeight * rows);

