package debug;

import org.apache.pdfbox.pdmodel.PDPageContentStream;
import org.apache.pdfbox.pdmodel.font.PDFont;

import java.awt.*;
import java.io.IOException;

/**
 * Implementers will be notified when vertical space is about to be consumed.
 * This class encapsulates PDPageContentStream to enable operations for height
 * calculations while contentStream effectively is null, and thus calling
 * write-functions will not really write to the PDF.
 */
public class PdfTextWriter {

    private final PDPageContentStream contentStream;

    private PDFont font;
    private float fontSize;
    private float textLineHeight;

    /**
     * Creates a new PdfTextWriter working with given optional contentStream.
     * @param contentStream can be null for just calculating a page count.
     */
    public PdfTextWriter(PDPageContentStream contentStream, PDFont font, float fontSize) throws IOException {
        assert font != null && fontSize > 0f;

        this.contentStream = contentStream;
        setFont(font, fontSize);
    }

    /**
     * Called after having moved to a new vertical location. Does nothing.
     * Override this to listen to vertical advance.
     * @param newHeight the new location to be taken.
     */
    public void advancedVertically(float newHeight, Object rowItem) {
    }

    public final void setFont(PDFont font, float fontSize) throws IOException {
        this.font = font;
        this.fontSize = fontSize;
        this.textLineHeight = PdfBoxUtil.lineHeight(font, fontSize);

        if (contentStream != null)
            contentStream.setFont(font, fontSize);
    }
    public void setFont(PDFont font) throws IOException {
        setFont(font, getFontSize());
    }

    public PDFont getFont() {
        return font;
    }
    public float getFontSize() {
        return fontSize;
    }
    public float getFontHeight() throws IOException {
        return PdfBoxUtil.fontHeight(getFont(), getFontSize());
    }
    public float getTextLineHeight() {
        return textLineHeight;
    }
    public float getStringWidth(String singleLineText) throws IOException {
        return PdfBoxUtil.stringWidth(getFont(), getFontSize(), singleLineText);
    }

    public void beginText() throws IOException {
        if (contentStream != null)
            contentStream.beginText();
    }
    public void newLineAtOffset(float startX, float startY) throws IOException {
        if (contentStream != null)
            contentStream.newLineAtOffset(startX, startY);
    }
    public void showText(String line) throws IOException {
        if (contentStream != null)
            contentStream.showText(line);
    }
    public void endText() throws IOException {
        if (contentStream != null)
            contentStream.endText();
    }
    public void close() throws IOException {
        if (contentStream != null)
            contentStream.close();
    }
    public void moveTo(float xStart, float yStart) throws IOException {
        if (contentStream != null)
            contentStream.moveTo(xStart, yStart);
    }
    public void lineTo(float xEnd, float yEnd) throws IOException {
        if (contentStream != null)
            contentStream.lineTo(xEnd, yEnd);
    }
    public void stroke() throws IOException {
        if (contentStream != null)
            contentStream.stroke();
    }
    public void setStrokingColor(Color color) throws IOException {
        if (contentStream != null)
            contentStream.setStrokingColor(color);
    }
    public void setLineWidth(float lineThickness) throws IOException {
        if (contentStream != null)
            contentStream.setLineWidth(lineThickness);
    }
}
