package fri.util.pdfbox.model;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Iterator;

import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.pdmodel.common.PDRectangle;

import fri.util.pdfbox.model.geometry.Dimension;
import fri.util.pdfbox.model.geometry.Padding;
import fri.util.pdfbox.model.geometry.Point;

/**
 * Conveniences and defaults around a PDFBox document.
 * 
 * @author fri May 5, 2019
 */
public class PdfBoxDocument implements Iterable<PDPage>
{
    public static final PDRectangle DEFAULT_PAGESIZE = PDRectangle.A4;
    
    public final PDDocument document;
    
    private final PDRectangle pageSize;
    private final Padding padding;
    
    private PdfBoxPageWriter writer;
    
    public PdfBoxDocument() {
        this(null);
    }
    
    public PdfBoxDocument(PDRectangle pageSize) {
       this(null, pageSize);
    }
    
    public PdfBoxDocument(PDDocument document, PDRectangle pageSize)  {
        this(document, pageSize, null);
    }
    
    public PdfBoxDocument(PDDocument document, PDRectangle pageSize, Padding padding)  {
        this.document = (document != null) ? document : new PDDocument();
        this.pageSize = (pageSize != null) ? pageSize : DEFAULT_PAGESIZE;
        this.padding = (padding != null) ? padding : new Padding(0, 0);
    }
    
    
    public int getPageCount()   {
        return document.getNumberOfPages();
    }
    
    public PDPage getPage(int index0ToN)   {
        return document.getPage(index0ToN);
    }
    
    /** @return the last page, or null if none was added ever. */
    public PDPage getLastPage() {
        return document.getNumberOfPages() > 0 ? document.getPage(document.getNumberOfPages() - 1) : null;
    }
    
    public int indexOf(PDPage page)   {
        for (int i = 0; i < getPageCount(); i++)
            if (document.getPage(i) == page)
                return i;
        return -1;
    }
    
    public PDPage removePage(int index0ToN)   {
        final PDPage page = document.getPage(index0ToN);
        document.removePage(index0ToN);
        return page;
    }
    
    public void removePage(PDPage page)   {
        document.removePage(page);
    }

    @Override
    public Iterator<PDPage> iterator() {
        return document.getPages().iterator();
    }
    
    public PDPage addPage(PDRectangle pageSize) throws IOException {
        final PDPage page = newPage(pageSize);
        document.addPage(page);
        return page;
    }
    
    public PDPage addPage() throws IOException {
        PDPage page = getLastPage();
        return addPage(page != null ? page.getMediaBox() : pageSize);
    }
    
    public PDPage insertPage(int index0ToN, PDRectangle pageSize) throws IOException {
        if (index0ToN >= getPageCount())
            return addPage(pageSize != null ? pageSize : pageSize);
        
        final PDPage movedPage = getPage(index0ToN);
        final PDPage page = newPage(pageSize);
        document.getPages().insertBefore(page, movedPage);
        return page;
    }
    
    /**
     * Factory method.
     * Doesn't close the current writer!
     */
    protected PDPage newPage(PDRectangle pageSize) throws IOException {
        assert pageSize != null;
        PDPage page = new PDPage(pageSize);
        PDRectangle mediaBox = page.getMediaBox();
        
        PDRectangle trimBox = new PDRectangle(
                mediaBox.getLowerLeftX() + padding.left,
                mediaBox.getLowerLeftY() + padding.top,
                mediaBox.getWidth() - padding.left - padding.right,
                mediaBox.getHeight() - padding.top - padding.bottom
            );
        page.setTrimBox(trimBox);
        //System.err.println("trimBox: "+trimBox);
        
        return page;
    }
    
    
    public Dimension pageDimension() throws IOException {
        return getOrCreateWriter().pageDimension();
    }
    
    public Point pageOrigin() throws IOException {
        return getOrCreateWriter().pageOrigin();
    }
    
    
    /** @return the currently active page-content-stream writer, or a new one if none has been created yet. */
    public final PdfBoxPageWriter getOrCreateWriter() throws IOException    {
        return (getWriter() != null) ? getWriter() : newWriter();
    }

    /** @return the currently active page-content-stream writer, or null if none has been created yet. */
    private PdfBoxPageWriter getWriter() throws IOException    {
        return writer;
    }

    /** Adds a new page and returns a content stream to it. */
    public final PdfBoxPageWriter newWriter() throws IOException    {
        return newWriter(addPage());
    }

    /**
     * Returns a new content stream to given page, but doesn't add the page.
     * Doesn't close the current writer.
     */
    public final PdfBoxPageWriter newWriter(PDPage page) throws IOException    {
        return writer = newPdfBoxPageWriter(page);
    }
    
    protected PdfBoxPageWriter newPdfBoxPageWriter(PDPage page) throws IOException  {
        return new PdfBoxPageWriter(this, page);
    }

    
    public void closePage() throws IOException {
        if (getWriter() != null) {
            getWriter().close();    // is tolerant against multiple close calls
            writer = null;
        }
    }

    public void saveAndClose(String filePath) throws IOException {
        saveAndClose(filePath, null, null);
    }

    public void saveAndClose(File file) throws IOException {
        saveAndClose(null, file, null);
    }

    public void saveAndClose(OutputStream outputStream) throws IOException {
        saveAndClose(null, null, outputStream);
    }

    private void saveAndClose(String filePath, File file, OutputStream outputStream) throws IOException {
        closePage();
        
        if (outputStream != null)
            document.save(outputStream);
        else if (file != null)
            document.save(file);
        else if (filePath != null)
            document.save(filePath);
        
        document.close();
    }

}
