package fri.util.pdfbox.model;

import java.awt.Color;
import java.io.IOException;

import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.pdmodel.PDPageContentStream;
import org.apache.pdfbox.pdmodel.PDPageContentStream.AppendMode;
import org.apache.pdfbox.pdmodel.common.PDRectangle;

import fri.util.pdfbox.model.geometry.Dimension;
import fri.util.pdfbox.model.geometry.Font;
import fri.util.pdfbox.model.geometry.Point;
import fri.util.pdfbox.model.geometry.Rectangle;

public class PdfBoxPageWriter
{
    public final PDPage page;
    public final PdfBoxDocument document;
    public final PDRectangle box;
    
    private PDPageContentStream contentStream;
    private boolean active = true;
    
    public PdfBoxPageWriter(PdfBoxDocument document, PDPage page) throws IOException {
        assert document != null && page != null;
        
        this.document = document;
        this.page = page;
        this.box = page.getTrimBox();
    }
    
    public PDPageContentStream getContentStream() throws IOException  {
        if (contentStream == null)
            contentStream = newPDPageContentStream(document.document, page);
        return contentStream;
    }
    
    protected PDPageContentStream newPDPageContentStream(PDDocument document, PDPage page) throws IOException  {
        return new PDPageContentStream(document, page, AppendMode.APPEND, false);
    }

    public boolean isActive() {
        return active;
    }
    public void setActive(boolean active) {
        this.active = active;
    }
    
    public Dimension pageDimension()   {
        return new Dimension(box.getWidth(), box.getHeight());
    }
    
    /** @return top-left corner of the page's trim-box, with x increasing right and y downwards. */
    public Point pageOrigin()   {
        return new Point(box.getLowerLeftX(), box.getHeight() - box.getUpperRightY());
    }
    
    public void close() throws IOException {
        contentStream.close();
        contentStream = null;
    }


    public void drawLine(Point start, Point end, Color color) throws IOException {
        if (isActive())  {
            if (color != null)
                getContentStream().setStrokingColor(color);
            
            getContentStream().moveTo(toPdfX(start.x), toPdfY(start.y));
            getContentStream().lineTo(toPdfX(end.x), toPdfY(end.y));
            
            getContentStream().stroke();
            
            if (color != null)
                getContentStream().setStrokingColor(Color.BLACK);
        }
    }

    public void drawBar(Rectangle bar, Color color) throws IOException {
        drawRect(bar, true, color);
    }

    public void drawRectangle(Rectangle rectangle, Color color) throws IOException {
        drawRect(rectangle, false, color);
    }

    public void drawRect(Rectangle rectangle, boolean fill, Color color) throws IOException {
        if (isActive())  {
            if (color != null)
                getContentStream().setStrokingColor(color);
            
            getContentStream().addRect(toPdfX(rectangle.topLeft.x), toPdfY(rectangle.bottomRight.y), rectangle.width(), rectangle.height());
            
            if (fill)
                getContentStream().fill();
            
            getContentStream().stroke();
            
            if (color != null)
                getContentStream().setStrokingColor(Color.BLACK);
        }
    }

    
    /** Start a paragraph. */
    public void startPrintText(Point topLeft, Font font) throws IOException {
        if (isActive())  {
            getContentStream().beginText();
            getContentStream().setFont(font.font, font.fontSize);
            getContentStream().newLineAtOffset(toPdfX(topLeft.x), toPdfY(topLeft.y + font.baseLine));
        }
    }
    
    /** Write a line of a paragraph. */
    public void printTextAndNewLine(String line, Font font) throws IOException {
        if (isActive())  {
            getContentStream().showText(line);
            getContentStream().newLineAtOffset(0, -font.lineHeight);
        }
    }
    
    /** End a paragraph. */
    public void endPrintText() throws IOException {
        if (isActive())
            getContentStream().endText();
    }

    
    private float toPdfX(float x) {
        assert x >= 0f && x < pageDimension().width : "Drawing into invisible area, wrong x = "+x;
        return pageOrigin().x + x;
    }

    private float toPdfY(float y) {
        assert y >= 0f && y < pageDimension().height : "Drawing into invisible area, wrong y = "+y;
        return pageDimension().height - (y + pageOrigin().y);
    }

}
