package fri.util.pdfbox.model.components;

import java.io.IOException;
import java.util.List;

import org.apache.pdfbox.pdmodel.PDPage;

import fri.util.pdfbox.model.PdfBoxPageWriter;
import fri.util.pdfbox.model.geometry.Font;
import fri.util.pdfbox.model.geometry.Point;

/**
 * Can be broken across pages.
 * It calls an optional page-break listener while drawing.
 * Base-class for bringing content into geometric shapes like text-columns or tables.
 * 
 * @author fri May 4, 2019
 */
public abstract class PdfBoxMultiPageComponent extends PdfBoxComponent
{
    /**
     * A page break is a vertical action, meaning this listener
     * will be called when starting to write the next line of some text.
     * It won't be called when writing the next character.
     */
    public interface PageBreakListener
    {
        /**
         * Called before outputting a text line that may exceed page bounds.
         * @param nextX the x-coordinate that is about to be entered.
         * @param nextY the y-coordinate that is about to be entered.
         * @return null for no page-break intervention, else a new contentStream and y-coordinate.
         */
        PageBreak willAdvance(Point next) throws IOException;
    }
    
    public static class PageBreak
    {
        /** Is null when staying on page, else it is the stream of a new page. */
        public final PdfBoxPageWriter writer;
        
        /** The coordinate where to continue to draw. */
        public final Point whereToContinue;
        
        public PageBreak(Point whereToContinue, PdfBoxPageWriter writer) {
            this.whereToContinue = whereToContinue;
            this.writer = writer;
        }
    }
    
    
    private PageBreakListener pageBreakListener;

    /**
     * @param writer required, the PDF stream to write into.
     * @param font optional, the font to be used by any write-method.
     */
    protected PdfBoxMultiPageComponent(PdfBoxPageWriter writer, Font font) {
        super(writer, font);
    }
    
    protected final PageBreakListener getOrCreatePageBreakListener() {
        if (pageBreakListener == null)
            pageBreakListener = newPageBreakListener();
        return pageBreakListener;
    }
    
    /** Creates different kinds of page-break listeners. */
    protected abstract PageBreakListener newPageBreakListener();
    
    protected final boolean mustCallPageBreakListener() {
        return getWriter().isActive();
    }

    /**
     * Manages the page break listener result generically by setting
     * a new page-writer when one was created by the caller.
     * @return the point where to continue writing.
     */
    protected Point pageBroken(PageBreak pageBreak, List<PDPage> consumedPages) throws IOException {
        if (pageBreak.writer != null && pageBreak.writer != getWriter())    {   // another page has been opened
            setWriter(pageBreak.writer);    // closes the current writer
            consumedPages.add(pageBreak.writer.page);
        }
        // else: staying on same page
        
        // go to another position
        return pageBreak.whereToContinue;
    }

}
