package fri.util.pdfbox.model.components;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.pdfbox.pdmodel.PDPage;

import fri.util.pdfbox.model.PdfBoxDocument;
import fri.util.pdfbox.model.WriteState;
import fri.util.pdfbox.model.geometry.Font;
import fri.util.pdfbox.model.geometry.OpenRectangle;
import fri.util.pdfbox.model.geometry.Point;
import fri.util.pdfbox.model.geometry.Rectangle;

/**
 * Brings text into shapes like rectangles or tables.
 * Can be used for geometry calculations by passing a null <i>contentStream</i>.
 * 
 * @author fri May 4, 2019
 */
public class PdfBoxTable extends PdfBoxTextComponent
{
    /** Table layout. */
    public static class View
    {
        /** Column proportions, will be adapted to a given width or page width. */
        public final float[] columnWidths;
        /** Horizontal cell padding. */
        public final float cellPaddingX;
        /** Vertical cell padding. */
        public final float cellPaddingY;
        
        /**
         * @param columnWidths the widths of the table columns.
         * @param cellPaddingX the horizontal space around table cell contents.
         * @param cellPaddingY the vertical space around table cell contents.
         */
        public View(float[] columnWidths, float cellPaddingX, float cellPaddingY) {
            this.columnWidths = columnWidths;
            this.cellPaddingX = cellPaddingX;
            this.cellPaddingY = cellPaddingY;
        }
        
    }
    
    
    private final View tableView;
    
    /**
     * @param document required, the PDF document to write into.
     * @param font optional, the font to be used by any write-method, default is Helvetica-10.
     * @param tableView required, the view specification of the table.
     * @throws IOException when PDFBox throws one.
     */
    public PdfBoxTable(PdfBoxDocument document, Font font, View tableView) throws IOException {
        super(document.getOrCreateWriter(), font);
        assert tableView != null;
        this.tableView = tableView;
    }
    
    /**
     * Prints given text-matrix into an open-bottom rectangle at given left/upper coordinates.
     * Mind that it may grow out of page-space on bottom!
     * @param contents the texts to paint into table cells.
     * @param topLeft the left top coordinate of the rectangle where to print into, bigger than zero.
     * @param width optional, when bigger than zero, the width of the rectangle where to print into.
     * @return the coordinates where drawing stopped at bottom of the consumed vertical space.
     * @throws IOException when PdfBox throws one.
     */
    public WriteState draw(String[][] contents, OpenRectangle rectangle) throws IOException {
        assert contents.length > 0 && contents[0].length == tableView.columnWidths.length;
        
        final float width = pageAdjustedWidth(rectangle);
        final float[] tableColumnWidths = adaptColumnWidths(tableView.columnWidths, width);
        
        final List<PDPage> consumedPages = new ArrayList<>();
        consumedPages.add(getWriter().page);
        
        final int rows = contents.length;
        Point cursor = new Point(rectangle.topLeft);
        
        for (int row = 0; row < rows; row++)    {
            final List<List<String>> textColumns = toTextColumns(contents[row], tableColumnWidths, tableView.cellPaddingX);
            
            if (mustCallPageBreakListener())    {
                final PageBreak pageBreak = calculateTableRow(textColumns, cursor, tableColumnWidths);
                if (pageBreak != null) // on new page
                    cursor = pageBroken(pageBreak, consumedPages);
            }
            
            final float nextY = drawTableRow(textColumns, cursor, tableColumnWidths);
            
            cursor = new Point(rectangle.topLeft.x, nextY);
        }
        
        return new WriteState(
                cursor,
                new Rectangle(rectangle.topLeft, new Point(rectangle.topLeft.x + width, cursor.y)), 
                consumedPages.toArray(new PDPage[consumedPages.size()]));
    }

    
    private List<List<String>> toTextColumns(String[] tableRow, float[] columnWidths, float paddingX) throws IOException {
        final List<List<String>> textColumns = new ArrayList<>();
        for (int column = 0; column < columnWidths.length; column++)   {
            final float columnWidth = columnWidths[column];
            textColumns.add(new TextColumn(getFont()).splitToLines(tableRow[column], columnWidth - 2 * paddingX));
        }
        return textColumns;
    }
    
    private PageBreak calculateTableRow(List<List<String>> textColumns, Point topLeft, float[] tableColumnWidths) throws IOException {
        getWriter().setActive(false);   // switch off to do just height calculation
        try {
            final float nextY = drawTableRow(textColumns, topLeft, tableColumnWidths);
            return getOrCreatePageBreakListener().willAdvance(new Point(topLeft.x, nextY));
        }
        finally {
            getWriter().setActive(true);
        }
    }
    
    private float drawTableRow(List<List<String>> textColumns, Point topLeft, float[] tableColumnWidths) throws IOException {
        final int columns = textColumns.size();
        float maxY = 0f;
        float x = topLeft.x;
        for (int column = 0; column < columns; column++)    {
            final float columnWidth = tableColumnWidths[column];
            final WriteState writeState = drawTextColumn(textColumns.get(column), new Point(x + tableView.cellPaddingX, topLeft.y), columnWidth);
            final float nextY = writeState.consumedRectangle.bottomRight.y;
            x += columnWidth;
            
            if (nextY > maxY)
                maxY = nextY;
        }
        return maxY + tableView.cellPaddingY;
    }

    /**
     * Calculates new column widths to match given table width proportionally.
     * Does NOT modify given column widths!
     * @param columnWidths the intended widths of columns of the target table.
     * @param tableWidth the entire width of the target table.
     * @return new array containing corrected column widths.
     */
    private float[] adaptColumnWidths(float[] columnWidths, float tableWidth) {
        float sum = 0.0f;
        for (int i = 0; i < columnWidths.length; i++)
            sum += columnWidths[i];

        final float proportion = tableWidth / sum;

        float[] newColumnWidths = new float[columnWidths.length];
        for (int i = 0; i < columnWidths.length; i++)
            newColumnWidths[i] = proportion * columnWidths[i];

        return newColumnWidths;
    }


}
