package fri.util.pdfbox.model.components;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.pdfbox.pdmodel.PDPage;

import fri.util.pdfbox.model.PdfBoxPageWriter;
import fri.util.pdfbox.model.WriteState;
import fri.util.pdfbox.model.geometry.Font;
import fri.util.pdfbox.model.geometry.OpenRectangle;
import fri.util.pdfbox.model.geometry.Point;
import fri.util.pdfbox.model.geometry.Rectangle;
import fri.util.pdfbox.model.pagination.DefaultPageBreakListener;

/**
 * Can print text as open-bottom multiline columns, or as single line.
 * 
 * @author fri May 4, 2019
 */
public class PdfBoxTextComponent extends PdfBoxMultiPageComponent
{
    /**
     * @param font optional, the font to be used by any write-method.
     * @param writer optional, the PDF stream to write into.
     */
    protected PdfBoxTextComponent(PdfBoxPageWriter writer, Font font) {
        super(writer, font);
    }
    
    /**
     * Prints given text as a single line at given left/upper coordinates.
     * Mind that this may grow out of page-space on right side!
     * @param text the text to draw.
     * @param topLeft the left coordinate of the text.
     * @return the y-coordinate where drawing stopped, i.e. the bottom of the consumed vertical space.
     * @throws IOException when PdfBox throws one.
     */
    public WriteState drawText(String text, Point topLeft) throws IOException {
        return drawTextColumn(text, new OpenRectangle(topLeft, Float.MAX_VALUE));
    }
    
    /**
     * Prints given text into an open-bottom rectangle starting at given left/upper coordinates.
     * This will not grow out of page-space on right side, but it may grow to next page on bottom,
     * so make sure a <code>PageBreakListener</code> was installed.
     * @param text the text to draw.
     * @param rectangle the open-bottom rectangle where to print into.
     * @return the y-coordinate where drawing stopped, i.e. the bottom of the consumed vertical space.
     * @throws IOException when PdfBox throws one.
     */
    public WriteState drawTextColumn(String text, OpenRectangle rectangle) throws IOException {
        final float width = pageAdjustedWidth(rectangle);
        return drawTextColumn(new TextColumn(getFont()).splitToLines(text, width), rectangle.topLeft, width);
    }
    
    protected WriteState drawTextColumn(List<String> textColumn, Point topLeft, float width) throws IOException {
        final float lineHeight = getFont().lineHeight;
        
        final List<PDPage> consumedPages = new ArrayList<>();
        consumedPages.add(getWriter().page);
        
        Point cursor = new Point(topLeft);
        float maxY = cursor.y;
        Point lastCursor = cursor;
        final String lastLine = (textColumn.size() > 0) ? textColumn.get(textColumn.size() - 1) : null;
        
        getWriter().startPrintText(cursor, getFont());
        
        for (String line : textColumn)    {
            if (mustCallPageBreakListener())    {
                final PageBreak pageBreak = getOrCreatePageBreakListener().willAdvance(new Point(cursor.x, cursor.y + lineHeight));
                if (pageBreak != null) { // on new page
                    getWriter().endPrintText();
                    final Point whereToContinue = pageBroken(pageBreak, consumedPages);
                    final float x = (whereToContinue.x < topLeft.x) ? topLeft.x : whereToContinue.x;
                    final float y = (whereToContinue.y < topLeft.y && consumedPages.size() == 1) ? topLeft.y : whereToContinue.y;
                    cursor = new Point(x, y);
                    maxY = cursor.y;
                    getWriter().startPrintText(cursor, getFont());
                }
            }
            
            getWriter().printTextAndNewLine(line, getFont());
            
            if (line == lastLine)    {
                float lineWidth = (line.trim().length() > 0) ? getFont().stringWidth(line+" ") : 0f;
                lastCursor = new Point(cursor.x + lineWidth, cursor.y);
            }
            
            cursor = new Point(cursor.x, cursor.y + lineHeight);
            if (maxY < cursor.y)
                maxY = cursor.y;
        }
        
        getWriter().endPrintText();
        
        return new WriteState(
                    lastCursor,
                    new Rectangle(topLeft, new Point(cursor.x + width, maxY)),
                    consumedPages.toArray(new PDPage[consumedPages.size()])
                );
    }

    @Override
    protected PageBreakListener newPageBreakListener() {
        return new DefaultPageBreakListener(getWriter().document, getWriter().pageDimension().height);
    }

}
