package fri.util.pdfbox.model.components;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import fri.util.pdfbox.model.geometry.Font;

/**
 * Formats a text to fit into a given column width.
 * Uses spaces of any kind, dash '-' and underscore '_'
 * as word break separators where to set line breaks.
 * Puts line-breaks also into very long words (like URLs)
 * when no such separator was found and the text is too long.
 * Preserves additional newlines that mark paragraphs.
 * 
 * @author fri May 4, 2019
 */
public class TextColumn
{
    private final Font font;
    
    public TextColumn(Font font) {
        this.font = font;
    }
    
    public List<String> splitToLines(String text, final float width) throws IOException {
        text = text.replaceAll("\t", " "); // PdfBox doesn't accept tabs
        text = text.replaceAll("[ ]{2,}", " "); // normalize spaces
        
        String[] paragraphs = text.split("\r\n");   // WINDOWS
        if (paragraphs.length == 1)
            paragraphs = text.split("\n");  // UNIX, OSX
        
        final List<String> lines = new ArrayList<String>();
        for (String paragraph : paragraphs) // a paragraph is a chunk of text that has no newlines.
            for (String line : splitParagraphToFitWidth(paragraph, width))
                lines.add(line);
        
        return lines;
    }
    
    private List<String> splitParagraphToFitWidth(String text, final float width) throws IOException {
        final List<String> lines = new ArrayList<String>();
        if (text.length() <= 0) {   // preserve newlines
            lines.add(text);  // will serve as empty line
            return lines;
        }
        
        int previousSpaceIndex = -1;
        
        while (text.length() > 0) {
            int nextSpaceIndex = nextSpaceIndex(text, previousSpaceIndex + 1);  // search first space
            if (nextSpaceIndex < 0)
                nextSpaceIndex = text.length(); // no space found
            
            final String line = text.substring(0, nextSpaceIndex).trim();
            final float lineWidth = font.stringWidth(line);
            
            if (lineWidth > width) {   // must split at preceding space if any
                if (previousSpaceIndex < 0) // no preceding space, split by width
                    previousSpaceIndex = nextSplitIndex(text, width);
                
                lines.add(text.substring(0, previousSpaceIndex));
                text = text.substring(previousSpaceIndex).trim();
                previousSpaceIndex = -1;
            }
            else if (nextSpaceIndex == text.length()) { // reached end of text
                lines.add(line);
                text = "";
            }
            else    {
                previousSpaceIndex = nextSpaceIndex;
            }
        }
        
        return lines;
    }

    private int nextSpaceIndex(String text, final int searchStartIndex) {
        for (int i = searchStartIndex; i < text.length(); i++)
            if (Character.isWhitespace(text.charAt(i)))
                return i;
        return -1;
    }

    private int nextSplitIndex(String text, float width) throws IOException {
        width -= font.fontSize * 0.39f;   // empiric correction
        for (int i = 1; i < text.length(); i++) {
            final String line = text.substring(0, i).trim();
            final float lineWidth = font.stringWidth(line);
            if (lineWidth >= width)
                return i;
        }
        return text.length();
    }

}
