package pdfboxtest;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.pdmodel.PDPageContentStream;
import org.apache.pdfbox.pdmodel.common.PDRectangle;
import org.apache.pdfbox.pdmodel.font.PDFont;
import org.apache.pdfbox.pdmodel.font.PDType1Font;

public class MultilineText
{
    public static void main(String[] args) throws IOException {
        new MultilineText().print();
    }
    
    private String text = 
        "By default when you write a (long) text using PDPageContentStream#showText() it is printed on a single line. "+
        "We need to calculate how many words will fit on a single line and then write the text to the document. "+
        "Here comes a newline:"+
        "\n"+
        "Here comes a long text without spaces: "+
        "Loremipsumdolorsitametconsecteturadipiscingelitseddoeiusmodtemporincididuntutlaboreetdoloremagnaaliqua. "+
        "\n"+
        "Several newlines are preserved:"+
        "\n"+
        "\n"+
        "Redundant        spaces         get removed,       this is called normalizing. Here comes text with tabs: "+
        "laboris\tnisi\tut\taliquip\tex\tea\tcommodo\tconsequat.\tDuis\taute\tirure\tdolor\tin\treprehenderit\tin\tut "+
        "voluptate velit esse cillum dolore eu fugiat nulla pariatur. "+
        "http://longlonglongurlwithoutend/longlonglongurlwithoutend/longlonglongurlwithoutend";

    public void print() throws IOException {
        final PDDocument document = new PDDocument();
        
        final PDFont font = PDType1Font.HELVETICA;
        final float fontSize = 10;
        final float fontHeight = fontSize * font.getFontDescriptor().getCapHeight() / 1000;
        
        final PDRectangle pageSize = PDRectangle.A4;
        final PDPage page = new PDPage(pageSize);
        document.addPage(page);
        
        final PDRectangle pageBox = page.getMediaBox();
        final float width = pageBox.getWidth();
        final float height = pageBox.getHeight();
        
        final float lineHeight = fontHeight * 1.5f; // standard line height is 150% of font height
        float startX = pageBox.getLowerLeftX();
        float startY = pageBox.getUpperRightY();
        
        final PDPageContentStream contentStream = new PDPageContentStream(document, page);
        float textBlockWidth = 120f;
        float y = displayFlowingText(contentStream, text, textBlockWidth, startX, startY, lineHeight, font, fontSize);
        
        contentStream.setLineWidth(1f);
        contentStream.addRect(startX, y, textBlockWidth, startY - y);
        contentStream.stroke();
        contentStream.close();
        
        final String fileName = "MultilineText.pdf";
        document.save(fileName);
        document.close();
    }
    

    private float displayFlowingText(
            PDPageContentStream contentStream,
            String text,
            float width,
            float startX,
            float startY,
            float lineHeight,
            PDFont font,
            float fontSize) throws IOException
    {
        contentStream.beginText();
        contentStream.setFont(font, fontSize);
        contentStream.newLineAtOffset(startX, startY - lineHeight);
        for (String line : splitToFitWidth(text, width, font, fontSize))    {
            contentStream.showText(line);
            contentStream.newLineAtOffset(0, -lineHeight);
            startY -= lineHeight;
        }
        contentStream.endText();
        float fontDescent = fontSize * font.getFontDescriptor().getDescent() / 1000f;
        return startY + fontDescent;
    }
    
    private List<String> splitToFitWidth(String text, final float width, PDFont font, float fontSize) throws IOException {
        text = text.replaceAll("\t", " "); // PdfBox doesn't accept tabs
        text = text.replaceAll("[ ]{2,}", " "); // normalize spaces
        
        String[] paragraphs = text.split("\r\n");   // WINDOWS
        if (paragraphs.length == 1)
            paragraphs = text.split("\n");  // UNIX, OSX
        
        final List<String> lines = new ArrayList<String>();
        for (String paragraph : paragraphs)
            for (String line : splitParagraphToFitWidth(paragraph, width, font, fontSize))
                lines.add(line);
        
        return lines;
    }
    
    private List<String> splitParagraphToFitWidth(String text, final float width, PDFont font, float fontSize) throws IOException {
        final List<String> lines = new ArrayList<String>();
        if (text.length() <= 0) {   // preserve newlines
            lines.add("");
            return lines;
        }
        
        int previousSpaceIndex = -1;
        
        while (text.length() > 0) {
            int nextSpaceIndex = nextSpaceIndex(text, previousSpaceIndex + 1);  // search first space
            if (nextSpaceIndex < 0)
                nextSpaceIndex = text.length(); // no space found
            
            final String line = text.substring(0, nextSpaceIndex).trim();
            final float lineWidth = stringWidth(font, fontSize, line);
            
            if (lineWidth > width) {   // must split at preceding space if any
                if (previousSpaceIndex < 0) // no preceding space, split by width
                    previousSpaceIndex = nextSplitIndex(text, width, font, fontSize);
                
                lines.add(text.substring(0, previousSpaceIndex));
                text = text.substring(previousSpaceIndex).trim();
                previousSpaceIndex = -1;
            }
            else if (nextSpaceIndex == text.length()) { // reached end of text
                lines.add(line);
                text = "";
            }
            else    {
                previousSpaceIndex = nextSpaceIndex;
            }
        }
        
        return lines;
    }

    private int nextSpaceIndex(String text, int searchStartIndex) {
        for (int i = searchStartIndex; i < text.length(); i++)
            if (Character.isWhitespace(text.charAt(i)))
                return i;
        return -1;
    }

    private int nextSplitIndex(String text, float width, PDFont font, float fontSize) throws IOException {
        width -= fontSize * 0.39f;   // empiric correction
        for (int i = 1; i < text.length(); i++) {
            final String line = text.substring(0, i).trim();
            final float lineWidth = stringWidth(font, fontSize, line);
            if (lineWidth >= width)
                return i;
        }
        return text.length();
    }
    
    private float stringWidth(PDFont font, float fontSize, String text) throws IOException {
        return fontSize * font.getStringWidth(text) / 1000f;
    }

}
